#!/usr/bin/env python
# vim:fileencoding=UTF-8:ts=4:sw=4:sta:et:sts=4:ai
from __future__ import (unicode_literals, division, absolute_import,
                        print_function)

__license__   = 'GPL v3'
__copyright__ = '2012, David Forrester <davidfor@internode.on.net>'
__docformat__ = 'restructuredtext en'

import time, os

from contextlib import closing

from calibre.utils.ipc.server import Server
from calibre.utils.ipc.job import ParallelJob
from calibre.utils.logging import Log
from calibre.constants import DEBUG

from calibre_plugins.koboutilities.action import MIMETYPE_KOBO
#from calibre_plugins.koboutilities.common_utils import debug_print

BASE_TIME = None
def debug_print(*args):
    global BASE_TIME
    if BASE_TIME is None:
        BASE_TIME = time.time()
#    if DEBUG:
    print('DEBUG: %6.1f'%(time.time()-BASE_TIME), *args)

def do_koboutilitiesa(books_to_scan, options, cpus, notification=lambda x,y:x):
    '''
    Master job, to launch child jobs to modify each ePub
    '''
    server = Server(pool_size=cpus)

    print("do_koboutilities - options=%s" % (options))
    # Queue all the jobs

    args = ['calibre_plugins.koboutilities.jobs', 'do_koboutilities_single',
            (books_to_scan, options)]
    print("do_koboutilities - args=%s" % (args))
    job = ParallelJob('arbitrary', str(book_id), done=None, args=args)
    job._book_id    = book_id
    job._title      = title
    job._authors    = authors
    job._contentIDs = contentIDs
    server.add_job(job)

    for book_id, contentIDs, title, authors in books_to_scan:
        print("do_koboutilities - book_id=%s, title=%s, authors=%s" % (book_id, title, authors))
        args = ['calibre_plugins.koboutilities.jobs', 'do_koboutilities_single',
                (book_id, contentIDs, options)]
        print("do_koboutilities - args=%s" % (args))
        job = ParallelJob('arbitrary', str(book_id), done=None, args=args)
        job._book_id    = book_id
        job._title      = title
        job._authors    = authors
        job._contentIDs = contentIDs
        server.add_job(job)

    # This server is an arbitrary_n job, so there is a notifier available.
    # Set the % complete to a small number to avoid the 'unavailable' indicator
    notification(0.01, 'Storing reading locations')

    # dequeue the job results as they arrive, saving the results
    total = len(books_to_scan)
    count = 0
    stored_locations = dict()
    while True:
        job = server.changed_jobs_queue.get()
        # A job can 'change' when it is not finished, for example if it
        # produces a notification. Ignore these.
        job.update()
        if not job.is_finished:
            continue
        # A job really finished. Get the information.
        stored_location = job.result
        book_id = job._book_id
        stored_locations[book_id] = stored_location
        count += 1
        notification(float(count)/total, 'Storing locations')
        # Add this job's output to the current log
        print('Logfile for book ID %d (%s / %s)'%(book_id, job._title, job._authors))
        print(job.details)
        print("\tstored_location=", stored_location)
        if count >= total:
            # All done!
            break

    server.close()
    # return the map as the job result
    return stored_locations


def do_store_locations(books_to_scan, options, cpus, notification=lambda x,y:x):
    '''
    Master job, to launch child jobs to modify each ePub
    '''
    debug_print("do_koboutilities - start")
    server = Server(pool_size=cpus)

    debug_print("do_koboutilities - options=%s" % (options))
    # Queue all the jobs

    args = ['calibre_plugins.koboutilities.jobs', 'do_koboutilities_all',
            (books_to_scan, options)]
#    debug_print("do_koboutilities - args=%s" % (args))
    debug_print("do_koboutilities - len(books_to_scan)=%d" % (len(books_to_scan)))
    job = ParallelJob('arbitrary', "Store locations", done=None, args=args)
#    job._book_id    = book_id
#    job._title      = title
#    job._authors    = authors
#    job._contentIDs = contentIDs
    server.add_job(job)

    # This server is an arbitrary_n job, so there is a notifier available.
    # Set the % complete to a small number to avoid the 'unavailable' indicator
    notification(0.01, 'Reading device database')

    # dequeue the job results as they arrive, saving the results
    total = 1
    count = 0
    stored_locations = dict()
    while True:
        job = server.changed_jobs_queue.get()
        # A job can 'change' when it is not finished, for example if it
        # produces a notification. Ignore these.
        job.update()
        if not job.is_finished:
            continue
        # A job really finished. Get the information.
        stored_locations = job.result
#        book_id = job._book_id
#        stored_locations[book_id] = stored_location
        count += 1
        notification(float(count)/total, 'Storing locations')
        # Add this job's output to the current log
        #debug_print("Stored_location=", stored_locations)
        debug_print("Stored_location count=", len(stored_locations))
        debug_print(job.details)
        if count >= total:
            # All done!
            break

    server.close()
    debug_print("do_koboutilities - finished")
    # return the map as the job result
    return stored_locations


def do_koboutilities_single(book_id, contentIDs, options):
    '''
    Child job, to store location for this book
    '''
    return _store_current_bookmark(Log(), book_id, contentIDs, options)


def do_koboutilities_all(books, options):
    '''
    Child job, to store location for all the books
    '''
    return _store_bookmarks(Log(), books, options)


def _store_current_bookmark(log, book_id, contentIDs, options):

    count_books = 0
    result      = None

    import sqlite3 as sqlite
    with closing(sqlite.connect(options["device_database_path"])) as connection:
        # return bytestrings if the content cannot the decoded as unicode
        connection.text_factory = lambda x: unicode(x, "utf-8", "ignore")

        epub_fetch_query = 'SELECT c1.ChapterIDBookmarked, ' \
                                'c2.adobe_location, '      \
                                'c1.ReadStatus, '          \
                                'c1.___PercentRead, '      \
                                'c1.Attribution, '         \
                                'c1.DateLastRead, '        \
                                'c1.Title, '               \
                                'c1.MimeType, '            \
                                'r.rating, '               \
                                'c1.contentID '            \
                            'FROM content c1 JOIN content c2 ON c1.ChapterIDBookmarked = c2.ContentID ' \
                                'LEFT OUTER JOIN ratings r ON c1.ContentID = r.ContentID '  \
                            'WHERE c2.bookid = ?'

        kepub_fetch_query = 'SELECT c1.ChapterIDBookmarked, ' \
                                'c1.adobe_location, '      \
                                'c1.ReadStatus, '          \
                                'c1.___PercentRead, '      \
                                'c1.Attribution, '         \
                                'c1.DateLastRead, '        \
                                'c1.Title, '               \
                                'c1.MimeType, '            \
                                'r.rating, '               \
                                'c1.contentID '            \
                            'FROM content c1 LEFT OUTER JOIN ratings r ON c1.ContentID = r.ContentID '  \
                            'WHERE c1.contentID = ?'

        cursor = connection.cursor()
        count_books += 1
        
        for contentID in contentIDs:
            log("store_current_bookmark - contentId='%s'" % (contentID))
            fetch_values = (contentID,)
            if contentID.endswith(".kepub.epub"):
                fetch_query = kepub_fetch_query
            else:
                fetch_query = epub_fetch_query
            cursor.execute(fetch_query, fetch_values)
            result = cursor.fetchone()

        connection.commit()
        cursor.close()
    
    return result


def _store_bookmarks(log, books, options):

    debug_print("_store_bookmarks - start")
    count_books      = 0
    stored_locations = dict()

    import sqlite3 as sqlite
    with closing(sqlite.connect(options["device_database_path"])) as connection:
        # return bytestrings if the content cannot the decoded as unicode
        connection.text_factory = lambda x: unicode(x, "utf-8", "ignore")

        epub_fetch_query = 'SELECT c1.ChapterIDBookmarked, ' \
                                'c2.adobe_location, '      \
                                'c1.ReadStatus, '          \
                                'c1.___PercentRead, '      \
                                'c1.Attribution, '         \
                                'c1.DateLastRead, '        \
                                'c1.Title, '               \
                                'c1.MimeType, '            \
                                'r.rating, '               \
                                'c1.contentID '            \
                            'FROM content c1 JOIN content c2 ON c1.ChapterIDBookmarked = c2.ContentID ' \
                                'LEFT OUTER JOIN ratings r ON c1.ContentID = r.ContentID '  \
                            'WHERE c1.ContentID = ?'

        kepub_fetch_query = 'SELECT c1.ChapterIDBookmarked, ' \
                                'c1.adobe_location, '      \
                                'c1.ReadStatus, '          \
                                'c1.___PercentRead, '      \
                                'c1.Attribution, '         \
                                'c1.DateLastRead, '        \
                                'c1.Title, '               \
                                'c1.MimeType, '            \
                                'r.rating, '               \
                                'c1.contentID '            \
                            'FROM content c1 LEFT OUTER JOIN ratings r ON c1.ContentID = r.ContentID '  \
                            'WHERE c1.contentID = ?'

        cursor = connection.cursor()
        count_books += 1
        
        debug_print("_store_bookmarks - about to start book loop")
        for book_id, contentIDs, title, authors in books:
            result = None
            debug_print("_store_bookmarks - contentIds='%s'" % (contentIDs))
            for contentID in contentIDs:
#                log("_store_bookmarks - contentId='%s'" % (contentID))
                debug_print("_store_bookmarks - contentId='%s'" % (contentID))
                fetch_values = (contentID,)
                if contentID.endswith(".kepub.epub"):
                    fetch_query = kepub_fetch_query
                else:
                    fetch_query = epub_fetch_query
                cursor.execute(fetch_query, fetch_values)
                result = cursor.fetchone()
            stored_locations[book_id] = result

        debug_print("_store_bookmarks - finished book loop")
        connection.commit()
        cursor.close()
    
    debug_print("_store_bookmarks - finished")
    return stored_locations


def do_clean_images_dir(options, cpus, notification=lambda x,y:x):
    main_image_path      = options['main_image_path']
    sd_image_path        = options['sd_image_path']
    device_database_path = options["device_database_path"]

    notification(1/5, 'Getting ImageIDs from main images directory')
    imageids_files_main = _get_file_imageIds(main_image_path)

    notification(2/6, 'Getting ImageIDs from SD card images directory')
    imageids_files_sd   = _get_file_imageIds(sd_image_path)
    
    notification(3/6, 'Getting ImageIDs from device database.')
    imageids_db = _get_imageId_set(device_database_path)
#        debug_print("clean_images_dir - len(imageids_db)=%d imageids_db=%s" % (len(imageids_db), imageids_db))

    notification(4/7, 'Checking/removing images from main images directory')
    extra_imageids_files_main = imageids_files_main - imageids_db
    notification(5/7, 'Checking/removing images from SD card images directory')
    extra_imageids_files_sd   = imageids_files_sd   - imageids_db
    debug_print("clean_images_dir - len(extra_imageids_files_main)=%d" % (len(extra_imageids_files_main)))
    debug_print("clean_images_dir - len(imageids_files_main)=%d" % (len(imageids_files_main)))
#    debug_print("clean_images_dir - len(extra_imageids_files_main)=%d extra_imageids_files_main=%s" % (len(extra_imageids_files_main), extra_imageids_files_main))
#    debug_print("clean_images_dir - len(imageids_files_main)=%d imageids_files_main=%s" % (len(imageids_files_main), imageids_files_main))

    extra_image_files_main = _remove_extra_files(extra_imageids_files_main, options['delete_extra_covers'], main_image_path)
    extra_image_files_sd   = _remove_extra_files(extra_imageids_files_sd,   options['delete_extra_covers'], sd_image_path)

    extra_image_files = {}
    extra_image_files['main_memory'] = extra_image_files_main
    extra_image_files['sd_card']     = extra_image_files_sd

    notification(7/7, 'Cleaning images directory - Done')

    return extra_image_files

def _get_file_imageIds(image_path):
    imageids_files = []
    for path, dirs, files in os.walk(image_path):
        debug_print("_get_file_imageIds - path=%s, dirs=%s" % (path, dirs))
        for filename in files:
            if filename.find(" - N3_") > 0: #                    debug_print("check_covers - filename=%s" % (filename))
                imageid = filename.split(" - N3_")[0]
                imageids_files.append(imageid)
    
    imageids_files = set(imageids_files)
    return imageids_files

def _remove_extra_files(extra_imageids_files, delete_extra_covers, image_path):
    extra_image_files = []
    from glob import glob
    for imageId in extra_imageids_files:
        for filename in glob(os.path.join(image_path, imageId + '*')):
            debug_print("_remove_extra_files - filename=%s" % (filename))
            extra_image_files.append(os.path.basename(filename))
            if delete_extra_covers:
                os.unlink(filename)
    
    return extra_image_files

def _get_imageId_set(device_database_path):
    import sqlite3 as sqlite
    with closing(sqlite.connect(device_database_path)) as connection:
        connection.text_factory = lambda x: unicode(x, "utf-8", "ignore")

        imageId_query = 'SELECT DISTINCT ImageId '       \
                        'FROM content '         \
                        'WHERE ContentType = 6'
        cursor = connection.cursor()

        imageIDs = []
        cursor.execute(imageId_query)
        for i, row in enumerate(cursor):
            imageIDs.append(row[0])
#            debug_print("_get_imageid_set - row[0]='%s'" % (row[0]))
        connection.commit()

        cursor.close()

    return set(imageIDs)
